/*
 * Copyright (C) 2011-2015 Daniel Scharrer
 *
 * This software is provided 'as-is', without any express or implied
 * warranty.  In no event will the author(s) be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 */

/*!
 * \file
 *
 * Strings describing the innoextract version.
 */
#ifndef INNOEXTRACT_RELEASE_HPP
#define INNOEXTRACT_RELEASE_HPP

//! Name of the program being built
extern const char innoextract_name[];

//! Name + version of the program being built
extern const char innoextract_version[];

//! Range of supported Inno Setup versions
extern const char innosetup_versions[];

//! Bug tracker URL
extern const char innoextract_bugs[];

//! Copyright line for the current program
extern const char innoextract_copyright[];

//! License text for the current program
extern const char innoextract_license[];

#endif // INNOEXTRACT_RELEASE_HPP
